#include <stdio.h>
#include <stdlib.h>

#include "ovdisis.h"
#include "various.h"

static void vdi_v_bar(VDI_Workstation *);
static void vdi_v_arc(VDI_Workstation *);
static void vdi_v_pieslice(VDI_Workstation *);
static void vdi_v_circle(VDI_Workstation *);
static void vdi_v_ellipse(VDI_Workstation *);
static void vdi_v_ellarc(VDI_Workstation *);
static void vdi_v_ellpie(VDI_Workstation *);
static void vdi_v_rbox(VDI_Workstation *);
static void vdi_v_rfbox(VDI_Workstation *);
static void vdi_v_justified(VDI_Workstation *);


VdiFunction *vdi_gdpfuncs[] = {
  UNUSED, vdi_v_bar, vdi_v_arc, vdi_v_pieslice, vdi_v_circle, vdi_v_ellipse,
  vdi_v_ellarc, vdi_v_ellpie, vdi_v_rbox, vdi_v_rfbox, vdi_v_justified
};


void vdi_gdps(VDI_Workstation *vwk)
{
  /* Check if specified subroutine is listed in vwk->cangdps[] */

  if(vdi_gdpfuncs[vdipb->contrl[SUBROUTINE]] && 
     (vdipb->contrl[SUBROUTINE] <= 10) && (vdipb->contrl[SUBROUTINE] >= 0))
    vdi_gdpfuncs[vdipb->contrl[SUBROUTINE]](vwk);
  else
    EDEBUG("vdi_gdps: Subroutine %d doesn't exist!\n",vdipb->contrl[SUBROUTINE]);
}


void vdi_v_bar(VDI_Workstation *vwk)
{
  int y,ni;
  unsigned long col;
  RECT cor;

  cor.x1 = vdipb->ptsin[0];
  cor.y1 = vdipb->ptsin[1];
  cor.x2 = vdipb->ptsin[2];
  cor.y2 = vdipb->ptsin[3];

  if(do_rectclip(&cor, &vwk->clip))
    {
      ni = gem2tos_color(vwk->inq.attr.planes, vwk->fill_a.color);
      col = get_color(vwk, ni);

      switch(vwk->fill_a.interior)
	{
	case FIS_HOLLOW:
	  /* Just draw the frame. Maybe. */
	  break;
	case FIS_SOLID:
	  for(y=cor.y1 ; y<=cor.y2 ; y++)
	    FBhline(vwk->fb, cor.x1, cor.x2, y, col);
	  break;
	case FIS_PATTERN:
	case FIS_HATCH:
	case FIS_USER:
	  EDEBUG("v_bar: Unimplemented filling style, %d!\n", vwk->fill_a.interior);
	  break;
	default:
	  EDEBUG("v_bar: Unknown filling style, %d!\n", vwk->fill_a.interior);
	  break;
	}

      /* Don't draw frame if FIS_SOLID */
      if(vwk->fill_a.perimeter && (vwk->fill_a.interior!=FIS_SOLID))
	{
	  RECT lin;

	  lin.x1 = vdipb->ptsin[0];
	  lin.x2 = vdipb->ptsin[2];
	  lin.y1 = lin.y2 = vdipb->ptsin[1];
	  if(do_lineclip(&lin, &vwk->clip))
	    FBhline(vwk->fb, lin.x1, lin.x2, lin.y1, col);
	  lin.x1 = vdipb->ptsin[0];
	  lin.x2 = vdipb->ptsin[2];
	  lin.y1 = lin.y2 = vdipb->ptsin[3];
	  if(do_lineclip(&lin, &vwk->clip))
	    FBhline(vwk->fb, lin.x1, lin.x2, lin.y2, col);
	  lin.x1 = lin.x2 = vdipb->ptsin[0];
	  lin.y1 = vdipb->ptsin[1];
	  lin.y2 = vdipb->ptsin[3];
	  if(do_lineclip(&lin, &vwk->clip))
	    FBline(vwk->fb, lin.x1, lin.y1, lin.x1, lin.y2, col);
	  lin.x1 = lin.x2 = vdipb->ptsin[2];
	  lin.y1 = vdipb->ptsin[1];
	  lin.y2 = vdipb->ptsin[3];
	  if(do_lineclip(&lin, &vwk->clip))
	    FBline(vwk->fb, lin.x2, lin.y1, lin.x2, lin.y2, col);
	}
    }
}

void vdi_v_arc(VDI_Workstation *vwk)
{
  EDEBUG("v_arc: Call not implemented!\n");
}

void vdi_v_pieslice(VDI_Workstation *vwk)
{
  EDEBUG("v_pieslice: Call not implemented!\n");
}


/* Uses Bresenham's Circle Algorithm */
/* Good for whole circles, but I guess I'll
   have to come up with something better
   for arcs and pieslices */
void vdi_v_circle(VDI_Workstation *vwk)
{
  int radius,x_center,y_center;
  int p,x,y,ni;
  int *x_values;
  unsigned long col;
  RECT lin_u, lin_l;

  x_center = vdipb->ptsin[0];
  y_center = vdipb->ptsin[1];
  radius = vdipb->ptsin[4];

  IDEBUG("v_circle: x_center: %d, y_center: %d, radius: %d\n",x_center,y_center,radius);

  ni = gem2tos_color(vwk->inq.attr.planes, vwk->fill_a.color);
  col = get_color(vwk, ni);

  switch(vwk->fill_a.interior)
    {
    case FIS_HOLLOW:
      /* Just draw the frame. Maybe. */
      break;
    case FIS_SOLID:
      x_values = calloc(radius+1, sizeof(int));
      if(!x_values)
	{
	  EDEBUG("v_circle: Not enough memory to allocate circle buffers!\n");
	  return;
	}

      x = 0;
      y = radius;
      p = 3-2*radius;
      while(x < y)
	{
	  /* add the points to a buffer
	     which is used for filling */
	  list_circle_points(x, y, radius, x_values);
	  if(p < 0)
	    p += 4*x+6;
	  else
	    {
	      p += 4*(x-y)+10;
	      y--;
	    }
	  x++;
	}
      if(x == y)
	list_circle_points(x, y, radius, x_values);

      /* Draw horizontal lines for each line 
	 calculated previously using the
	 x_values list */
      /* Lines are mirrored around the x-axis
         and then mirrored around the y-axis */
      for(y=-radius ; y<0 ; y++)
	{
	  lin_l.x1 = lin_u.x1 = -x_values[y+radius] + x_center;
	  lin_l.x2 = lin_u.x2 =  x_values[y+radius] + x_center;
	  lin_u.y1 = lin_u.y2 =  y + y_center;
	  lin_l.y1 = lin_l.y2 = -y + y_center;
	  if(do_lineclip(&lin_u, &vwk->clip))
	    FBhline(vwk->fb, lin_u.x1, lin_u.x2, lin_u.y1, col);
	  if(do_lineclip(&lin_l, &vwk->clip))
	    FBhline(vwk->fb, lin_l.x1, lin_l.x2, lin_l.y1, col);
	}
      /* Finally the middle line which hasn't been drawn */
      lin_l.x1 = -x_values[radius] + x_center;
      lin_l.x2 =  x_values[radius] + x_center;
      lin_l.y1 = lin_l.y2 = y_center;
      if(do_lineclip(&lin_l, &vwk->clip))
	FBhline(vwk->fb, lin_l.x1, lin_l.x2, lin_l.y1, col);
 
      free(x_values);

     break;
    case FIS_PATTERN:
    case FIS_HATCH:
    case FIS_USER:
      EDEBUG("v_circle: Unimplemented filling style, %d!\n", vwk->fill_a.interior);
      break;
    default:
      EDEBUG("v_circle: Unknown filling style, %d!\n", vwk->fill_a.interior);
      break;
    }

  /* Don't draw frame if FIS_SOLID */
  if(vwk->fill_a.perimeter && (vwk->fill_a.interior!=FIS_SOLID))
    {
      x = 0;
      y = radius;
      p = 3-2*radius;

      while(x < y)
	{
	  set_circle_pixels(vwk->fb, x, y, x_center, y_center, &vwk->clip, col);
	  if(p < 0)
	    p += 4*x + 6;
	  else
	    {
	      p += 4*(x-y) + 10;
	      y--;
	    }
	  x++;
	}
      if(x == y)
	set_circle_pixels(vwk->fb, x, y, x_center, y_center, &vwk->clip, col);
    }
}

void vdi_v_ellipse(VDI_Workstation *vwk)
{
  EDEBUG("v_ellipse: Call not implemented!\n");
}

void vdi_v_ellarc(VDI_Workstation *vwk)
{
  EDEBUG("v_ellarc: Call not implemented!\n");
}

void vdi_v_ellpie(VDI_Workstation *vwk)
{
  EDEBUG("v_ellpie: Call not implemented!\n");
}

void vdi_v_rbox(VDI_Workstation *vwk)
{
  EDEBUG("v_rbox: Call not implemented!\n");
}

void vdi_v_rfbox(VDI_Workstation *vwk)
{
  EDEBUG("v_rfbox: Call not implemented!\n");
}

void vdi_v_justified(VDI_Workstation *vwk)
{
  EDEBUG("v_justified: Call not implemented!\n");
}


